/**************************************************************
*Super Awesome Not So Tiny Power Meter
*		using ATtiny85
*	by Bill Porter  
*	www.billporter.info
*
*
*
*  This program is free software: you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the Free 
*   Software Foundation, either version 3 of the License, or (at your option)
*   any later version.
*  This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   <http://www.gnu.org/licenses/>
*
*
*
***************************************************************/

/*
ATtiny has one port, portB
5 usable pins

Pin usage
PB0 		 Serial TX
PB1
PB2/ADC1	 Amp 1
PB3/ADC3	 Amp 2
PB4/ADC2	 Volts


The plan is to track time with Timer0 at 1/8 scale 
and expanded to 16 bit in software
At 1Mhz, i timer tick is .000001 seconds

18000 ticks is over a full period



*/

//includes
#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <inttypes.h>
#define F_CPU 8000000UL  //8Mhz
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

//Macros
#define bitRead(value, bit) (((value) >> (bit)) & 0x01)
#define bitSet(value, bit) ((value) |= (1UL << (bit)))
#define bitClear(value, bit) ((value) &= ~(1UL << (bit)))
#define bitWrite(value, bit, bitvalue) (bitvalue ? bitSet(value, bit) : bitClear(value, bit))

#define RMS(val) (float)val*.707

//reads hybrid timer, returns value
#define readTimer() (timerHighByte << 8) | TCNT0

//sets ADC mux to pin
#define setAnalogPin(pin) ADMUX = pin & 0b00000011

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif


//Some settings
#define baud 4800
#define bit_delay 1000000/baud

#define HIGH 0x1
#define LOW 0x0

#define TXPIN 0
#define AMPS1 1
#define AMPS2 3
#define VOLTS 2

//I like byte
typedef uint8_t byte;

//Global Variable
volatile byte timerHighByte;

//Function Protos
void loop();
int analogRead();
void digitalWrite(byte, byte);
void serialWrite(byte);
long map(long, long, long, long, long);


int main(){

//setup timer
sbi(TCCR0B, CS01); //prescale at 1/8
sbi(TIMSK, TOIE0); //enable overflow vector, used to extend timer to 16 bit

//enable interupts
sei();

//set pin directions
DDRB=0b00000001;

//clear pullups
PORTB=0b00000001;

//enable watchdog timer, 500ms delay before reset
wdt_enable(5);

for(;;){

	wdt_reset(); //reset watchdog
	
	loop();
	_delay_ms(200);
	}
}


//the main loop
void loop(){

	//setup some variables

	unsigned int max_volts = 0;
	unsigned int max_amps1 = 0;
	unsigned int max_amps2 = 0;

	unsigned int timing_max = 0;

	unsigned int volts1=0;
	unsigned int volts2=0;

	unsigned int amps1_peak = 0;
	unsigned int amps2_peak = 0;

	unsigned int freq = 0;
	unsigned int amps1_pf = 0;
	unsigned int amps2_pf = 0;

	unsigned int temp=0;
	//int check=0;

	char string[50];
	byte string_len = 0;
	byte CS=0;

	//configure ADC
	//set ADC clock at 125kHz
	sbi(ADCSRA, ADPS2);
	sbi(ADCSRA, ADPS1);
	cbi(ADCSRA, ADPS0);

	// enable a2d conversions
	sbi(ADCSRA, ADEN);

	//300 samples of each ADC pin will capture just over 2 period's worth
	//of waveform data. I'm just looking for the max value.
	//I came by 300 by testing code executioin speed on 8 Mhz ATtiny

	setAnalogPin(VOLTS);
	for(int i=0; i<=300; i++)
		if(temp=analogRead(), temp > max_volts)  max_volts = temp;

	setAnalogPin(AMPS1);
	for(int i=0; i<=300; i++)
		if(temp=analogRead(), temp > max_amps1)  max_amps1 = temp;

	setAnalogPin(AMPS2);
	for(int i=0; i<=300; i++)
		if(temp=analogRead(), temp > max_amps2)  max_amps2 = temp;

	//now time to capture some timming data. 
	//first we are going to reconfigure ADC for faster reads, at the cost of accuracy
	//though we are only looking for when it peaks, so who cares about accuracy.

	// disable a2d conversions
	cbi(ADCSRA, ADEN);

	//set ADC clock at 250KHz

	sbi(ADCSRA, ADPS2);
	cbi(ADCSRA, ADPS1);
	sbi(ADCSRA, ADPS0);

	// enable ADC
	sbi(ADCSRA, ADEN);

	setAnalogPin(VOLTS);
	while(analogRead() != 0) {} //wait for a valley
	while(analogRead() == 0) {} //wait for the start

	for(int i=0; i<200; i++){
		if(temp=analogRead(), temp > timing_max) {
			 timing_max = temp;
			 volts1 = readTimer();
		}
	}

	timing_max = 0;

	while(analogRead() != 0) {} //wait for a valley
	while(analogRead() == 0) {} //wait for the start

	for(int i=0; i<200; i++){
		if(temp=analogRead(), temp > timing_max) {
			 timing_max = temp;
			 volts2 = readTimer();
		}
	}

	//record period time
	freq = volts2 - volts1;

	 //fix the delays in reading ADC. Found experimentally comparing to a proper o-scope. 
	freq -= (freq / 34); 

	//lets do this again for pf

	if(max_amps1 != 0)	{
		timing_max = 0;

		while(analogRead() != 0) {} //wait for a valley
		while(analogRead() == 0) {} //wait for the start

		for(int i=0; i<200; i++){
			if(temp=analogRead(), temp > timing_max) {
				 timing_max = temp;
				 volts1 = readTimer();
			}
		}

		setAnalogPin(AMPS1);
		timing_max = 0;

		while(analogRead() != 0) {} //wait for a valley
		while(analogRead() == 0) {} //wait for the start

		for(int i=0; i<200; i++){
			if(temp=analogRead(), temp > timing_max) {
				 timing_max = temp;
				 amps1_peak = readTimer();
			}
		}

		//store the difference in time
		amps1_peak -= volts1;
	
		//fix the delays in reading ADC. Found experimentally comparing to a proper o-scope. 
		amps1_peak -= (amps1_peak / 34); 
		//if the difference is larger then one cycle, divide by # of cycles it's bigger
		//to reduce the sampel to < 1 cycle
		amps1_peak -= ((amps1_peak / freq) * freq);

		temp = freq/2;
	
		//get within 1/2 cycle if larger
		if(amps1_peak > temp) amps1_peak -=temp;

		// 1/4 cycle is the time when pf = 0
		temp /= 2;
	
		if(amps1_peak < temp)  //handles different mapping for leading/trailing pf
			amps1_pf = map(amps1_peak, temp, 0, 0, 100); //scale length of time to PF
		else
			amps1_pf = map(amps1_peak, freq/2, temp, 100, 0); //scale length of time to PF

	}


	//rinse and repeat
	if(max_amps2 != 0)	{
		
		timing_max = 0;

		setAnalogPin(VOLTS);
		while(analogRead() != 0) {} //wait for a valley
		while(analogRead() == 0) {} //wait for the start

		for(int i=0; i<200; i++){
			if(temp=analogRead(), temp > timing_max) {
				 timing_max = temp;
				 volts1 = readTimer();
			}
		}

		setAnalogPin(AMPS2);
		timing_max = 0;

		while(analogRead() != 0) {} //wait for a valley
		while(analogRead() == 0) {} //wait for the start

		for(int i=0; i<200; i++){
			if(temp=analogRead(), temp > timing_max) {
				 timing_max = temp;
				 amps2_peak = readTimer();
			}
		}

		//store the difference in time
		amps2_peak -= volts1;
	
		//fix the delays in reading ADC. Found experimentally comparing to a proper o-scope. 
		amps2_peak -= (amps2_peak / 34); 
		//if the difference is larger then one cycle, divide by # of cycles it's bigger
		//to reduce the sampel to < 1 cycle
		amps2_peak -= ((amps2_peak / freq) * freq);

		temp = freq/2;
	
		//get within 1/2 cycle if larger
		if(amps2_peak > temp) amps2_peak -=temp;

		// 1/4 cycle is the time when pf = 0
		temp /= 2;
	
		if(amps2_peak < temp)  //handles different mapping for leading/trailing pf
			amps2_pf = map(amps2_peak, temp, 0, 0, 100); //scale length of time to PF
		else
			amps2_pf = map(amps2_peak, freq/2, temp, 100, 0); //scale length of time to PF

	}

	
	//measurments done. We have everything we need. To to start some conversions
	

	// disable a2d conversions
	cbi(ADCSRA, ADEN);
	
	//remap ADC values to actual voltages, in millivolts. 5 volts = 5000 millivolts. 
	//we do that to avoid floating point, and make it easy to print later. 
	max_volts = map(max_volts, 0, 1023, 0, 4960);
	max_amps1 = map(max_amps1, 0, 1023, 0, 4960);
	max_amps2 = map(max_amps2, 0, 1023, 0, 4960);

	//upscale voltage to account for divider and Xformer
	max_volts = ((float)max_volts*42)/100;

	////convert peak to RMS
	max_volts = RMS(max_volts);
	max_amps1 = RMS(max_amps1);
	max_amps2 = RMS(max_amps2);  

	//measured transfer function, each one is different for the slight differences in OpAmp gain
	if(max_amps1 != 0) max_amps1 = (float)max_amps1 * .508 - 39;
	if(max_amps2 != 0) max_amps2 = (float)max_amps2 * .518 - 36;
	
	


	//time to send the data. Split into two runs or else it explodes. 
	//first we'll send a long header to get the RX's attention. 

	for(int i=0; i<20; i++)
		serialWrite('X');

	serialWrite('$');

	string_len = sprintf(string, "POWER,%u.%u,%lu.%lu,%u.%u,", max_volts/10, max_volts%10, (10000000/freq)/10, (10000000/freq)%10, max_amps1/10, max_amps1%10);

	for(int i = 0; i< string_len; i++){
		serialWrite(string[i]);
		CS^=string[i];
		}
	
	string_len = sprintf(string, "%u.%u,.%02u,.%02u", max_amps2/10, max_amps2%10, amps1_pf, amps2_pf);
	//string_len = sprintf(string, "Samples: %u \n", check);    

	for(int i = 0; i< string_len; i++){
		serialWrite(string[i]);
		CS^=string[i];
		}

	serialWrite('*');

	sprintf(string, "%X", CS);

	for(int i = 0; i< 2; i++)
		serialWrite(string[i]);

	serialWrite('\n');



}



//timer 0 overflow
SIGNAL(TIM0_OVF_vect){

//extend timer to 16 bit here. 
	timerHighByte++;

}

int analogRead() {

	uint8_t low, high;

	// start the conversion
	sbi(ADCSRA, ADSC);

	// ADSC is cleared when the conversion finishes
	while (bit_is_set(ADCSRA, ADSC));

	// we have to read ADCL first; doing so locks both ADCL
	// and ADCH until ADCH is read.  reading ADCL second would
	// cause the results of each conversion to be discarded,
	// as ADCL and ADCH would be locked when it completed.
	low = ADCL;
	high = ADCH;

	// combine the two bytes
	return (high << 8) | low;
}


//write out a byte as software emulated Uart
void serialWrite(byte bite){

digitalWrite(TXPIN, LOW);  //signal start bit
_delay_us(bit_delay);

 	for (byte mask = 0x01; mask; mask <<= 1) {
	    if (bite & mask){ // choose bit
	      digitalWrite(TXPIN,HIGH); // send 1
	    }
	    else{
	      digitalWrite(TXPIN,LOW); // send 0
	    }
	    _delay_us(bit_delay);
	  } 

	

	digitalWrite(TXPIN, HIGH);  //signal end bit
	_delay_us(bit_delay);

}

void digitalWrite(byte pin, byte val){

	byte bit = _BV(pin);

	if(val == LOW) PORTB &= ~bit;
	else PORTB |= bit;
}



long map(long x, long in_min, long in_max, long out_min, long out_max)
{
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}
